from spynnaker.pyNN.utilities.random_stats import AbstractRandomStats

from scipy.stats import uniform


class RandomStatsUniformImpl(AbstractRandomStats):
    """ An implementation of AbstractRandomStats for uniform distributions\
        (as scipy.stats.uniform takes slightly different parameters to\
        numpy.random.uniform)
    """

    def _get_params(self, dist):
        return [dist.parameters[0], dist.parameters[1] - dist.parameters[0]]

    def cdf(self, dist, v):
        return uniform.cdf(v, *self._get_params(dist))

    def ppf(self, dist, p):
        return uniform.ppf(p, *self._get_params(dist))

    def mean(self, dist):
        return uniform.mean(*self._get_params(dist))

    def std(self, dist):
        return uniform.std(*self._get_params(dist))

    def var(self, dist):
        return uniform.var(*self._get_params(dist))

    def high(self, dist):
        """ Return the variance of the distribution
        """
        if dist.boundaries is None:
            return None
        else:
            return max(dist.boundaries)

    def low(self, dist):
        """ Return the variance of the distribution
        """
        if dist.boundaries is None:
            return None
        else:
            return min(dist.boundaries)



from spynnaker.pyNN.models.pynn_population_common import PyNNPopulationCommon
from spynnaker.pyNN.models.recording_common import RecordingCommon
from spynnaker.pyNN.utilities import utility_calls
from spinn_front_end_common.utilities import globals_variables
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from pyNN import descriptions

import numpy
import logging

logger = logging.getLogger(__name__)


class Population(PyNNPopulationCommon, RecordingCommon):
    """ A collection neuron of the same types. It encapsulates a type of\
        vertex used with Spiking Neural Networks, comprising n cells (atoms)\
        of the same model type.

    :param int size:
        size (number of cells) of the Population.
    :param cellclass:
        specifies the neural model to use for the Population
    :param dict cellparams:
        a dictionary containing model specific parameters and values
    :param structure:
        a spatial structure
    :param string label:
        a label identifying the Population
    """

    def __init__(self, size, cellclass, cellparams, spinnaker, label,
                 structure=None):

        internal_cellparams = dict(cellparams)

        # set spinnaker targeted parameters
        model_label = None
        if 'label' in internal_cellparams:
            model_label = internal_cellparams['label']
        internal_cellparams['label'] = self.create_label(model_label, label)
        internal_cellparams['n_neurons'] = size

        # create population vertex.
        vertex = cellclass(**internal_cellparams)

        PyNNPopulationCommon.__init__(
            self, spinnaker_control=spinnaker, size=size, vertex=vertex,
            initial_values=None, structure=structure)
        RecordingCommon.__init__(
            self, self, globals_variables.get_simulator().machine_time_step)

    @property
    def default_parameters(self):
        """ The default parameters of the vertex from this population
        """
        return self._vertex.default_parameters

    def describe(self, template='population_default.txt', engine='default'):
        """ Returns a human-readable description of the population.

        The output may be customised by specifying a different template
        together with an associated template engine (see ``pyNN.descriptions``)

        If template is None, then a dictionary containing the template context
        will be returned.
        """

        vertex_context = self._vertex.describe()

        context = {
            "label": self.label,
            "celltype": descriptions.render(
                engine, 'modeltype_default.txt', vertex_context),
            "structure": None,
            "size": self.size,
            "size_local": self.size,
            "first_id": None,
            "last_id": None,
        }
        if self.size > 0:
            context.update({
                "local_first_id": None,
                "cell_parameters": {}})

        if self._structure:
            context["structure"] = self._structure.describe(template=None)
        return descriptions.render(engine, template, context)

    # noinspection PyPep8Naming
    def getSpikes(self, compatible_output=True, gather=True):
        """
        Return a 2-column numpy array containing cell ids and spike times for\
        recorded cells.
        """

        self._compatible_output_and_gather_warnings(compatible_output, gather)
        return self._get_recorded_variable("spikes")

    def get_spike_counts(self, gather=True):
        """ Return the number of spikes for each neuron.
        """
        spikes = self.getSpikes(True, gather)
        n_spikes = {}
        counts = numpy.bincount(spikes[:, 0].astype(dtype=numpy.int32),
                                minlength=self._vertex.n_atoms)
        for i in range(self._vertex.n_atoms):
            n_spikes[i] = counts[i]
        return n_spikes

    # noinspection PyUnusedLocal
    def get_gsyn(self, gather=True, compatible_output=True):
        """
        Return a 3-column numpy array containing cell ids, time and synaptic\
        conductances for recorded cells.

        :param gather: not used - inserted to match PyNN specs
        :type gather: bool
        :param compatible_output: not used - inserted to match PyNN specs
        :type compatible_output: bool
        """

        self._compatible_output_and_gather_warnings(compatible_output, gather)
        excit = self._get_recorded_variable("gsyn_exc")
        inhib = self._get_recorded_variable("gsyn_inh")
        # TODO this needs fixing for seperate recordings of cells on views
        # and assembliers
        # merge two arrays into one
        merged = numpy.hstack((excit, inhib))
        return numpy.delete(merged, [3, 4], 1)

    # noinspection PyUnusedLocal
    def get_v(self, gather=True, compatible_output=True):
        """
        Return a 3-column numpy array containing cell ids, time, and V_m for\
        recorded cells.

        :param gather: not used - inserted to match PyNN specs
        :type gather: bool
        :param compatible_output: not used - inserted to match PyNN specs
        :type compatible_output: bool
        """
        self._compatible_output_and_gather_warnings(compatible_output, gather)
        return self._get_recorded_variable("v")

    @staticmethod
    def _compatible_output_and_gather_warnings(compatible_output, gather):
        """ checks the values for compatible out and gather warnings

        :param compatible_output: if compatible with pynn
        :param gather: if gathering from pynn
        :return: None
        """
        if not gather:
            logger.warn(
                "Spynnaker 0.7 only supports gather = True, will  execute "
                "as if gather was true anyhow")

        if not compatible_output:
            logger.warn(
                "Spynnaker 0.7 only supports compatible_output = True, will"
                " execute as if compatible_output was false anyhow")

    @staticmethod
    def is_local(cell_id):
        """ Determine whether the cell with the given ID exists on the local \
            MPI node.

        :param cell_id:
        """

        # Doesn't really mean anything on SpiNNaker
        return True

    # noinspection PyPep8Naming
    def meanSpikeCount(self, gather=True):
        """ The mean number of spikes per neuron

        :param gather: gather has no meaning in spinnaker, always set to true
        :return: an array which contains the average spike rate per neuron
        """
        return self.mean_spike_count(gather)

    def mean_spike_count(self, gather=True):
        """ The mean number of spikes per neuron
        """
        spike_counts = self.get_spike_counts(gather)
        total_spikes = sum(spike_counts.values())
        return total_spikes / self._size

    def nearest(self, position):
        """ Return the neuron closest to the specified position
        """
        # doesn't always work correctly if a position is equidistant between
        # two neurons, i.e. 0.5 should be rounded up, but it isn't always.
        # also doesn't take account of periodic boundary conditions

        # TODO: Enable when __getitem__ is enabled
        # pos = numpy.array([position] * self.positions.shape[1]).transpose()
        # dist_arr = (self.positions - pos) ** 2
        # distances = dist_arr.sum(axis=0)
        # nearest = distances.argmin()
        # return self[nearest]

        raise NotImplementedError

    # noinspection PyPep8Naming
    def randomInit(self, distribution):
        """ Set initial membrane potentials for all the cells in the\
            population to random values.

        :param `pyNN.random.RandomDistribution` distribution:\
            the distribution used to draw random values.

        """
        self.initialize('v', distribution)
        self._change_requires_mapping = True

    def record(self, to_file=None):
        """ Record spikes from all cells in the Population.

        :param to_file: file to write the spike data to
        """

        self._record('spikes', self._create_full_filter_list(1), 1, to_file)

        # state that something has changed in the population,
        self._change_requires_mapping = True

    def record_gsyn(self, to_file=None):
        """ Record the synaptic conductance for all cells in the Population.

        :param to_file: the file to write the recorded gsyn to.
        """

        # have to set each to record and set the file at that point, otherwise
        # itll not work due to pynn bug
        self._record('gsyn_exc', self._create_full_filter_list(1), 1, to_file)
        self.file = to_file
        self._record('gsyn_inh', self._create_full_filter_list(1), 1, to_file)
        self.file = to_file

        # state that something has changed in the population,
        self._change_requires_mapping = True

    def record_v(self, to_file=None):
        """ Record the membrane potential for all cells in the Population.

        :param to_file: the file to write the recorded v to.
        """

        self._record('v', self._create_full_filter_list(1), 1, to_file)
        self.file = to_file

        # state that something has changed in the population,
        self._change_requires_mapping = True

    @property
    def positions(self):
        """ Return the position array for structured populations.
        """
        if self._positions is None:
            if self._structure is None:
                raise ValueError("attempted to retrieve positions "
                                 "for an unstructured population")
            self._positions = self._structure.generate_positions(
                self._vertex.n_atoms)
        return self._positions

    @positions.setter
    def positions(self, positions):
        """ Sets all the positions in the population.
        """
        self._positions = positions

        # state that something has changed in the population,
        self._change_requires_mapping = True

    # noinspection PyPep8Naming
    def printSpikes(self, filename, gather=True):
        """ Write spike time information from the population to a given file.

        :param filename: the absolute file path for where the spikes are to\
                    be printed in
        :param gather: Supported from the PyNN language, but ignored here
        """
        if not gather:
            logger.warn("Spynnaker only supports gather = true, will execute"
                        " as if gather was true anyhow")
        spikes = self._get_recorded_variable('spikes')
        if spikes is not None:
            first_id = 0
            num_neurons = self._vertex.n_atoms
            dimensions = self._vertex.n_atoms
            last_id = self._vertex.n_atoms - 1
            utility_calls.check_directory_exists_and_create_if_not(filename)
            spike_file = open(filename, "w")
            spike_file.write("# first_id = {}\n".format(first_id))
            spike_file.write("# n = {}\n".format(num_neurons))
            spike_file.write("# dimensions = [{}]\n".format(dimensions))
            spike_file.write("# last_id = {}\n".format(last_id))
            for (neuronId, time) in spikes:
                spike_file.write("{}\t{}\n".format(time, neuronId))
            spike_file.close()

    def print_gsyn(self, filename, gather=True):
        """ Write conductance information from the population to a given file.

        :param filename: the absolute file path for where the gsyn are to be\
                    printed in
        :param gather: Supported from the PyNN language, but ignored here
        """
        time_step = (self._spinnaker_control.machine_time_step * 1.0) / 1000.0
        gsyn_exc = self._get_recorded_variable('gsyn_exc')
        gsyn_inh = self._get_recorded_variable('gsyn_inh')

        first_id = 0
        num_neurons = self._vertex.n_atoms
        dimensions = self._vertex.n_atoms
        utility_calls.check_directory_exists_and_create_if_not(filename)
        file_handle = open(filename, "w")
        file_handle.write("# first_id = {}\n".format(first_id))
        file_handle.write("# n = {}\n".format(num_neurons))
        file_handle.write("# dt = {}\n".format(time_step))
        file_handle.write("# dimensions = [{}]\n".format(dimensions))
        file_handle.write("# last_id = {{}}\n".format(num_neurons - 1))
        file_handle = open(filename, "w")
        # TODO will need adjusting when filters and views assemblies work
        for ((neuronId, time, value_e), (_, _, value_i)) in zip(
                gsyn_exc, gsyn_inh):
            file_handle.write("{}\t{}\t{}\t{}\n".format(
                time, neuronId, value_e, value_i))
        file_handle.close()

    def print_v(self, filename, gather=True):
        """ Write membrane potential information from the population to a\
            given file.

        :param filename: the absolute file path for where the voltage are to\
                     be printed in
        :param gather: Supported from the PyNN language, but ignored here
        """
        time_step = (self._spinnaker_control.machine_time_step * 1.0) / 1000.0
        v = self._get_recorded_variable("v")
        utility_calls.check_directory_exists_and_create_if_not(filename)
        file_handle = open(filename, "w")
        first_id = 0
        num_neurons = self._vertex.n_atoms
        dimensions = self._vertex.n_atoms
        file_handle.write("# first_id = {}\n".format(first_id))
        file_handle.write("# n = {}\n".format(num_neurons))
        file_handle.write("# dt = {}\n".format(time_step))
        file_handle.write("# dimensions = [{}]\n".format(dimensions))
        file_handle.write("# last_id = {}\n".format(num_neurons - 1))
        for (neuronId, time, value) in v:
            file_handle.write("{}\t{}\t{}\n".format(time, neuronId, value))
        file_handle.close()

    def rset(self, parametername, rand_distr):
        """ 'Random' set. Set the value of parametername to a value taken\
             from rand_distr, which should be a RandomDistribution object.

        :param parametername: the parameter to set
        :param rand_distr: the random distribution object to set the parameter\
                     to
        """
        self.set(parametername, rand_distr)

    def sample(self, n, rng=None):
        """ Return a random selection of neurons from a population in the form\
            of a population view

        :param n: the number of neurons to sample
        :param rng: the random number generator to use.
        """

        # TODO: Need PopulationView support
        raise NotImplementedError

    def save_positions(self, file):  # @ReservedAssignment
        """ Save positions to file.

            :param file: the file to write the positions to.
        """
        file_handle = open(file, "w")
        file_handle.write(self.positions)
        file_handle.close()

    @property
    def structure(self):
        """ Return the structure for the population.
        """
        return self._structure

    def tset(self, parametername, value_array):
        """ 'Topographic' set. Set the value of parametername to the values in\
            value_array, which must have the same dimensions as the Population.

        :param parametername: the name of the parameter
        :param value_array: the array of values which must have the correct\
                number of elements.
        """
        if len(value_array) != self._vertex.n_atoms:
            raise ConfigurationException(
                "To use tset, you must have a array of values which matches "
                "the size of the population. Please change this and try "
                "again, or alternatively, use set()")
        self.set(parametername, value_array)

    def _end(self):
        """ Do final steps at the end of the simulation
        """
        if self._record_spike_file is not None:
            self.printSpikes(self._record_spike_file)
        if self._record_v_file is not None:
            self.print_v(self._record_v_file)
        if self._record_gsyn_file is not None:
            self.print_gsyn(self._record_gsyn_file)



from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence.\
    timing_dependence_spike_pair import \
    TimingDependenceSpikePair as CommonTimingDependenaceSpikePair


class TimingDependenceSpikePair(CommonTimingDependenaceSpikePair):

    def __init__(self, tau_plus=20.0, tau_minus=20.0):
        CommonTimingDependenaceSpikePair.__init__(
            self, tau_plus=tau_plus, tau_minus=tau_minus)



from spynnaker.pyNN.models.neuron.plasticity.stdp.weight_dependence.\
    weight_dependence_additive import WeightDependenceAdditive as \
    CommonWeightDependenceAdditive


class WeightDependenceAdditive(CommonWeightDependenceAdditive):

    # noinspection PyPep8Naming
    def __init__(
            self, w_min=0.0, w_max=1.0, A_plus=0.01, A_minus=0.01):
        CommonWeightDependenceAdditive.__init__(self, w_min=w_min, w_max=w_max)
        self.set_a_plus_a_minus(a_plus=A_plus, a_minus=A_minus)



import inspect as __inspect
import logging as __logging
import os as __os

import numpy as __numpy

import spynnaker7
from pyNN.random import NumpyRNG, RandomDistribution
from pyNN.space import \
    distance, Space, Line, Grid2D, Grid3D, Cuboid, Sphere, RandomStructure
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities import globals_variables

from spynnaker.pyNN.models.neural_projections \
    .delay_afferent_application_edge import DelayAfferentApplicationEdge
from spynnaker.pyNN.models.neural_projections.projection_application_edge \
    import ProjectionApplicationEdge
from spynnaker.pyNN.models.neuron.builds.if_cond_exp_base \
    import IFCondExpBase as IF_cond_exp
from spynnaker.pyNN.models.neuron.builds.if_curr_exp_base \
    import IFCurrExpBase as IF_curr_exp
from spynnaker.pyNN.models.neuron.synapse_dynamics.pynn_synapse_dynamics \
    import PyNNSynapseDynamics as SynapseDynamics
from spynnaker.pyNN.models.neuron.synapse_dynamics.synapse_dynamics_stdp \
    import SynapseDynamicsSTDP as STDPMechanism
from spynnaker.pyNN.models.spike_source.spike_source_array \
    import SpikeSourceArray
from spynnaker.pyNN.models.spike_source.spike_source_from_file \
    import SpikeSourceFromFile
from spynnaker.pyNN.models.spike_source.spike_source_poisson \
    import SpikeSourcePoisson
from spynnaker.pyNN.models.utility_models.delay_extension_vertex \
    import DelayExtensionVertex
from spynnaker.pyNN.utilities import utility_calls

from spynnaker7.pyNN.models.connectors.all_to_all_connector \
    import AllToAllConnector
from spynnaker7.pyNN.models.connectors. \
    distance_dependent_probability_connector import \
    DistanceDependentProbabilityConnector
from spynnaker7.pyNN.models.connectors. \
    fixed_number_post_connector import FixedNumberPostConnector
from spynnaker7.pyNN.models.connectors. \
    fixed_number_pre_connector import FixedNumberPreConnector
from spynnaker7.pyNN.models.connectors. \
    fixed_probability_connector import FixedProbabilityConnector
from spynnaker7.pyNN.models.connectors.from_file_connector \
    import FromFileConnector
from spynnaker7.pyNN.models.connectors.from_list_connector import \
    FromListConnector
from spynnaker7.pyNN.models.connectors.multapse_connector \
    import MultapseConnector
from spynnaker7.pyNN.models.connectors.one_to_one_connector \
    import OneToOneConnector
from spynnaker7.pyNN.models.plasticity_components.timing_dependence \
    .timing_dependence_spike_pair \
    import TimingDependenceSpikePair as SpikePairRule
from spynnaker7.pyNN.models.plasticity_components.weight_dependence.\
    weight_dependence_additive \
    import WeightDependenceAdditive as AdditiveWeightDependence
from spynnaker7.pyNN.models.plasticity_components.weight_dependence \
    .weight_dependence_multiplicative \
    import WeightDependenceMultiplicative as MultiplicativeWeightDependence

from spynnaker7.pyNN import external_devices
from spynnaker7.pyNN import extra_models

from spynnaker7.pyNN.spinnaker import Spinnaker as __Spinnaker
from spynnaker7._version import __version__  # NOQA
from spynnaker7._version import __version_name__  # NOQA
from spynnaker7._version import __version_month__  # NOQA
from spynnaker7._version import __version_year__  # NOQA

# traditional logger
logger = __logging.getLogger(__name__)

# List of binary search paths
_binary_search_paths = []

__all__ = [
    # Ugly, but tests expect it
    'utility_calls',
    # Implementations of the neuroscience models
    'IF_cond_exp', 'IF_curr_exp',
    'DelayAfferentApplicationEdge', 'DelayExtensionVertex',
    'ProjectionApplicationEdge', 'SpikeSourcePoisson', 'SpikeSourceArray',
    'SpikeSourceFromFile', 'AllToAllConnector', 'FixedNumberPreConnector',
    'FixedProbabilityConnector', 'FromListConnector', 'FromFileConnector',
    'MultapseConnector', 'OneToOneConnector', 'FixedNumberPostConnector',
    'DistanceDependentProbabilityConnector', 'SynapseDynamics',
    'STDPMechanism', 'AdditiveWeightDependence', 'SpikePairRule',
    'MultiplicativeWeightDependence',
    # Stuff from pyNN.random
    'NumpyRNG', 'RandomDistribution',
    # Stuff from pyNN.space
    'distance', 'Space', 'Line', 'Grid2D', 'Grid3D', 'Cuboid', 'Sphere',
    'RandomStructure',
    # External devices and extra models
    'external_devices', 'extra_models',
    # Stuff that we define
    'end', 'setup', 'run', 'get_spynnaker',
    'num_processes', 'rank', 'reset', 'set_number_of_neurons_per_core',
    'Population', 'Projection',
    'NativeRNG', 'get_current_time', 'create', 'connect', 'get_time_step',
    'get_min_delay', 'get_max_delay', 'set', 'initialize', 'record',
    'record_v', 'record_gsyn', 'get_machine']


def end():
    """
    Do any necessary cleaning up before exiting.

    Unregisters the controller,
    prints any data recorded using the low-level API
    """
    globals_variables.get_simulator().stop()
    # _spinnaker = None


def get_spynnaker():
    """helper method for other plugins to add stuff to the graph

    :return: The current spinnaker API, or None if before setup or after end.
    """
    return globals_variables.get_simulator()


def num_processes():
    """ Return the number of MPI processes
       (not used for SpiNNaker, always returns 1)
    """
    return 1


def rank():
    """ Return the MPI rank of the current node. (not used for SpiNNaker,\
        always returns 0 - as this is the minimum rank suggesting the front\
        node)
    """
    return 0


def reset():
    """ Reset the time to zero, and start the clock.
    """
    globals_variables.get_not_running_simulator().reset()


def run(run_time=None):
    """ Run the simulation for run_time ms.

    :param run_time: simulation length (in ms)
    """
    globals_variables.get_simulator().run(run_time)
    return None


def setup(timestep=0.1, min_delay=None, max_delay=None, machine=None,
          database_socket_addresses=None, n_chips_required=None,
          **extra_params):
    """ Should be called at the very beginning of a script.
        extra_params contains any keyword arguments that are required by a\
        given simulator but not by others.

    :param machine: A SpiNNaker machine used to run the simulation.
    :param timestep: The timestep in milleseconds.\
       Value will be rounded up to whole microseconds.\
       Set to None to use the value from the config file
    :param min_delay: the minumum number of time steps supported for delays
    :param max_delay: the maximum number of time steps supported for delays
    :param machine: The machine ip address
    :param database_socket_addresses: the set of sockets needed to be listened
    to for database notification protocol
    :param n_chips_required: The number of chips required for the simulation
    :param extra_params: random other crap
    :rtype: float or None
    """
    global _binary_search_paths

    logger.info(
        "sPyNNaker (c) {} APT Group, University of Manchester".format(
            __version_year__))
    parent_dir = __os.path.split(__os.path.split(spynnaker7.__file__)[0])[0]
    logger.info(
        "Release version {}({}) - {} {}. Installed in folder {}".format(
            __version__, __version_name__, __version_month__, __version_year__,
            parent_dir))

    if len(extra_params) > 0:
        logger.warn("Extra params {} have been applied to the setup "
                    "command which we do not consider".format(extra_params))
    __Spinnaker(
        host_name=machine, timestep=timestep, min_delay=min_delay,
        max_delay=max_delay,
        database_socket_addresses=database_socket_addresses,
        n_chips_required=n_chips_required)

    # the PyNN API expects the MPI rank to be returned
    return rank()


def set_number_of_neurons_per_core(neuron_type, max_permitted):
    """ Sets a ceiling on the number of neurons of a given type that can be\
        placed on a single core.

    :param neuron_type: the neuron type that will have its max atoms set
    :param max_permitted: The max amount of atoms to be set
    :type neuron_type: The string reprensetation of the neuron type
    :type max_permitted: int
    :rtype: None
    """
    if not __inspect.isclass(neuron_type):
        if neuron_type in globals():
            neuron_type = globals()[neuron_type]
        else:
            raise Exception("Unknown Vertex Type {}".format(neuron_type))

    simulator = globals_variables.get_not_running_simulator()
    simulator.set_number_of_neurons_per_core(neuron_type, max_permitted)


# noinspection PyPep8Naming
def Population(size, cellclass, cellparams, structure=None, label=None):
    """ building a new pop

    :param size: n neurons
    :param cellclass: the neuron class that needs to be created
    :param cellparams: the params to put into the neuron model
    :param structure: ??????
    :param label: the human readable label
    :return: a new population object
    """

    globals_variables.get_simulator().verify_not_running()
    return globals_variables.get_not_running_simulator().create_population(
        size, cellclass, cellparams, structure, label)


# noinspection PyPep8Naming
def Projection(presynaptic_population, postsynaptic_population,
               connector, source=None, target='excitatory',
               synapse_dynamics=None, label=None, rng=None):
    """ builds a new projection object

    :param presynaptic_population: the source pop
    :param postsynaptic_population: the dest pop
    :param connector: the connector describing connecitivty
    :param source: ??????????
    :param target: type of synapse, exicitiatory or inhibitoary for example.
    :param synapse_dynamics: plasticity
    :param label: human readable label
    :param rng: random number generator if needed
    :return: a new Projection object
    """
    globals_variables.get_simulator().verify_not_running()
    return globals_variables.get_not_running_simulator().create_projection(
        presynaptic_population, postsynaptic_population, connector, source,
        target, synapse_dynamics, label, rng)


def NativeRNG(seed_value):
    """ Fixes the random number generator's seed
    :param seed_value:
    :return:
    """
    __numpy.random.seed(seed_value)


def get_current_time():
    """
    returns the machine time step defined in setup
    :return: the runtime currently
    """
    return globals_variables.get_simulator().get_current_time()


# =============================================================================
#  Low-level API for creating, connecting and recording from individual neurons
# =============================================================================

def create(cellclass, cellparams=None, n=1):
    """ Create n cells all of the same type.

    If n > 1, return a list of cell ids/references.
    If n==1, return just the single id.
    """
    if cellparams is None:
        cellparams = {}
    return Population(n, cellclass, cellparams)


def connect(source, target, weight=0.0, delay=None, synapse_type="excitatory",
            p=1, rng=None):
    """ Connect a source of spikes to a synaptic target.

    source and target can both be individual cells or lists of cells, in
    which case all possible connections are made with probability p, using
    either the random number generator supplied, or the default rng
    otherwise. Weights should be in nA or uS.
    """
    connector = FixedProbabilityConnector(
        p_connect=p, weights=weight, delays=delay)
    return Projection(source, target, connector, target=synapse_type, rng=rng)


def get_time_step():
    """ The timestep requested
    :return:
    """
    return globals_variables.get_simulator().machine_time_step


def get_min_delay():
    """ The minimum allowed synaptic delay.
    :return:
    """
    return globals_variables.get_simulator().min_delay


def get_max_delay():
    """ The maximum allowed synaptic delay.
    :return:
    """
    return globals_variables.get_simulator().max_delay


def set(cells, param, val=None):  # @ReservedAssignment
    """ Set one or more parameters of an individual cell or list of cells.

    param can be a dict, in which case val should not be supplied, or a string
    giving the parameter name, in which case val is the parameter value.
    """
    assert isinstance(cells, Population)
    cells.set(param, val)


def initialize(cells, variable, value):
    cells.initialize(variable, value)


def record(source, filename):
    """ Record spikes to a file. source should be a Population.
    """
    source.record(to_file=filename)


def record_v(source, filename):
    """ Record spikes to a file. source should be a Population.
    """
    source.record_v(to_file=filename)


def record_gsyn(source, filename):
    """ Record spikes to a file. source should be a Population.
    """
    source.record_gsyn(to_file=filename)


def get_machine():
    """ Get the spinnaker machine in use
    """
    if not globals_variables.has_simulator():
        raise ConfigurationException(
            "You currently have not ran setup, please do so before calling "
            "get_machine")
    return globals_variables.get_simulator().machine



from spynnaker.pyNN.models.neural_projections.connectors. \
    distance_dependent_probability_connector import \
    DistanceDependentProbabilityConnector as \
    CommonDistanceDependentProbabilityConnector
from pyNN.space import Space


class DistanceDependentProbabilityConnector(
        CommonDistanceDependentProbabilityConnector):
    """ Make connections using a distribution which varies with distance.
    """

    def __init__(
            self, d_expression, weights=0.0, delays=1,
            allow_self_connections=True, space=Space(),
            safe=True, verbose=False, n_connections=None):
        """

        :param `string` d_expression:
            the right-hand side of a valid python expression for
            probability, involving 'd', e.g. "exp(-abs(d))", or "d<3",
            that can be parsed by eval(), that computes the distance
            dependent distribution
        :param `bool` allow_self_connections:
            if the connector is used to connect a
            Population to itself, this flag determines whether a neuron is
            allowed to connect to itself, or only to other neurons in the
            Population.
        :param `pyNN.Space` space:
            a Space object, needed if you wish to specify distance-
            dependent weights or delays
        :param `int` n_connections:
            The number of efferent synaptic connections per neuron.
        :param safe: if True, check that weights and delays have valid values.
         If False, this check is skipped.
        :param `float` weights:
            may either be a float, a !RandomDistribution object, a list/
            1D array with at least as many items as connections to be
            created, or a distance dependence as per a d_expression. Units nA.
        :param `float` delays:  -- as `weights`. If `None`, all synaptic delays
            will be set to the global minimum delay.
        """

        CommonDistanceDependentProbabilityConnector.__init__(
            self, d_expression=d_expression,
            allow_self_connections=allow_self_connections, safe=safe,
            verbose=verbose, n_connections=n_connections)
        self.set_weights_and_delays(weights, delays)
        self.set_space(space)



from spynnaker.pyNN.models.neuron.plasticity.stdp.weight_dependence\
    import WeightDependenceAdditiveTriplet \
    as CommonWeightDependenceAdditiveTriplet


class WeightDependenceAdditiveTriplet(CommonWeightDependenceAdditiveTriplet):

    # noinspection PyPep8Naming
    def __init__(
            self, w_min=0.0, w_max=1.0, A_plus=0.01, A_minus=0.01,
            A3_plus=0.01, A3_minus=0.01):
        CommonWeightDependenceAdditiveTriplet.__init__(
            self, w_max=w_max, w_min=w_min, A3_plus=A3_plus,
            A3_minus=A3_minus)
        self.set_a_plus_a_minus(a_minus=A_minus, a_plus=A_plus)



from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spynnaker.pyNN.utilities.spynnaker_failed_state \
    import SpynnakerFailedState
from spynnaker7.pyNN.spynnaker7_simulator_interface \
    import Spynnaker7SimulatorInterface


class Spynnaker7FailedState(Spynnaker7SimulatorInterface,
                            SpynnakerFailedState):

    __slots__ = ()

    def create_population(self, size, cellclass, cellparams, structure,
                          label):
        raise ConfigurationException(
            "This call is only valid between setup and end/stop")

    def create_projection(self, presynaptic_population,
                          postsynaptic_population, connector, source,
                          target, synapse_dynamics, label, rng):
        raise ConfigurationException(
            "This call is only valid between setup and end/stop")



from spynnaker.pyNN.models.neuron.synapse_dynamics.synapse_dynamics_static \
    import SynapseDynamicsStatic
from spynnaker.pyNN.models.neuron.abstract_population_vertex \
    import AbstractPopulationVertex
from spynnaker.pyNN.models.pynn_projection_common import PyNNProjectionCommon

from spinn_front_end_common.utilities.exceptions import ConfigurationException

import logging

logger = logging.getLogger(__name__)


# noinspection PyProtectedMember
class Projection(PyNNProjectionCommon):
    """ A container for all the connections of a given type (same synapse type\
        and plasticity mechanisms) between two populations, together with\
        methods to set parameters of those connections, including of\
        plasticity mechanisms.
    """

    # noinspection PyUnusedLocal
    def __init__(
            self, presynaptic_population, postsynaptic_population, label,
            connector, spinnaker_control, machine_time_step, user_max_delay,
            timescale_factor, source=None, target='excitatory',
            synapse_dynamics=None, rng=None):

        synapse_dynamics_stdp = None
        if synapse_dynamics is None:
            synapse_dynamics_stdp = SynapseDynamicsStatic()
        else:
            synapse_dynamics_stdp = synapse_dynamics.slow

        PyNNProjectionCommon.__init__(
            self, spinnaker_control=spinnaker_control, connector=connector,
            synapse_dynamics_stdp=synapse_dynamics_stdp,
            target=target, pre_synaptic_population=presynaptic_population,
            post_synaptic_population=postsynaptic_population,
            rng=rng, machine_time_step=machine_time_step,
            user_max_delay=user_max_delay, label=label,
            time_scale_factor=spinnaker_control.time_scale_factor)

        if not isinstance(postsynaptic_population._get_vertex,
                          AbstractPopulationVertex):
            raise ConfigurationException(
                "postsynaptic population is not designed to receive"
                " synaptic projections")

    def describe(self, template='projection_default.txt', engine='default'):
        """ Return a human-readable description of the projection.

        The output may be customised by specifying a different template
        together with an associated template engine (see ``pyNN.descriptions``)

        If template is None, then a dictionary containing the template context
        will be returned.
        """
        # TODO
        raise NotImplementedError

    def __getitem__(self, i):
        """Return the `i`th connection within the Projection."""
        # TODO: Need to work out what is being returned
        raise NotImplementedError

    # noinspection PyPep8Naming
    def getSynapseDynamics(
            self, parameter_name, format='list',  # @ReservedAssignment
            gather=True):  # @UnusedVariable
        """ Get parameters of the dynamic synapses for all connections in this\
            Projection.
        :param parameter_name:
        :param format:
        :param gather:
        """

        if not gather:
            logger.warn("Spynnaker always gathers from every core.")

        fixed_value = self._synapse_information.synapse_dynamics.get_value(
            parameter_name)
        return self._get_synaptic_data(
            format == "list", None, [(parameter_name, fixed_value)])

    # noinspection PyPep8Naming
    def getWeights(
            self, format='list',  # @ReservedAssignment
            gather=True):  # @UnusedVariable
        """
        Get synaptic weights for all connections in this Projection.

        Possible formats are: a list of length equal to the number of
        connections in the projection, a 2D weight array (with NaN for
        non-existent connections). Note that for the array format, if there is
        more than connection between two cells, the summed weight will be
        given.
        :param format: the type of format to be returned (only support "list")
        :param gather: gather the weights from stuff. currently has no meaning\
                in spinnaker when set to false. Therefore is always true
        """
        logger.info("Getting weights from Projection {}".format(self._label))
        return self._get_synaptic_data(format == "list", ["weight"])

    # noinspection PyPep8Naming
    def getDelays(self, format='list', gather=True):  # @ReservedAssignment
        """
        Get synaptic delays for all connections in this Projection.

        Possible formats are: a list of length equal to the number of
        connections in the projection, a 2D delay array (with NaN for
        non-existent connections).
        """
        return self._get_synaptic_data(format == "list", ["weight"])

    def __len__(self):
        """ Return the total number of local connections.
        """

        # TODO: Need to work out what this means
        raise NotImplementedError

    # noinspection PyPep8Naming
    def printDelays(self, file_name, list_format='list', gather=True):
        """ Print synaptic weights to file. In the array format, zeros are\
            printed for non-existent connections.
        """
        # TODO:
        raise NotImplementedError

    # noinspection PyPep8Naming
    def printWeights(self, file_name, list_format='list', gather=True):
        """ Print synaptic weights to file. In the array format, zeros are\
            printed for non-existent connections.
        """
        # TODO:
        raise NotImplementedError

    # noinspection PyPep8Naming
    def randomizeWeights(self, rand_distr):
        """ Set weights to random values taken from rand_distr.
        """
        # TODO: Requires that the synapse list is not created proactively
        raise NotImplementedError

    # noinspection PyPep8Naming
    def randomizeDelays(self, rand_distr):
        """ Set delays to random values taken from rand_distr.
        """
        # TODO: Requires that the synapse list is not created proactively
        raise NotImplementedError

    # noinspection PyPep8Naming
    def randomizeSynapseDynamics(self, param, rand_distr):
        """ Set parameters of the synapse dynamics to values taken from\
            rand_distr
        """
        # TODO: Look at what this is randomising
        raise NotImplementedError

    # noinspection PyPep8Naming
    def saveConnections(self, file_name, gather=True, compatible_output=True):
        """ Save connections to file in a format suitable for reading in with\
            a FromFileConnector.
        """
        # TODO
        raise NotImplementedError

    # noinspection PyPep8Naming
    def setDelays(self, d):
        """ Set the delays

        d can be a single number, in which case all delays are set to this\
        value, or a list/1D array of length equal to the number of connections\
        in the projection, or a 2D array with the same dimensions as the\
        connectivity matrix (as returned by `getDelays(format='array')`).
        """
        # TODO: Requires that the synapse list is not created proactively
        raise NotImplementedError

    # noinspection PyPep8Naming
    def setSynapseDynamics(self, param, value):
        """ Set parameters of the dynamic synapses for all connections in this\
            projection.
        """
        # TODO: Need to set this in the edge
        raise NotImplementedError

    # noinspection PyPep8Naming
    def setWeights(self, w):
        """ Set the weights

        w can be a single number, in which case all weights are set to this\
        value, or a list/1D array of length equal to the number of connections\
        in the projection, or a 2D array with the same dimensions as the\
        connectivity matrix (as returned by `getWeights(format='array')`).\
        Weights should be in nA for current-based and uS for conductance-based\
        synapses.
        """

        # TODO: Requires that the synapse list is not created proactively
        raise NotImplementedError

    # noinspection PyPep8Naming
    def weightHistogram(self, min_weight=None, max_weight=None, nbins=10):
        """ Return a histogram of synaptic weights.

        If min and max are not given, the minimum and maximum weights are\
        calculated automatically.
        """
        # TODO
        raise NotImplementedError



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase
from spinn_utilities.abstract_base import abstractmethod

from spynnaker.pyNN.spynnaker_simulator_interface \
    import SpynnakerSimulatorInterface


@add_metaclass(AbstractBase)
class Spynnaker7SimulatorInterface(SpynnakerSimulatorInterface):

    __slots__ = ()

    @abstractmethod
    def create_population(self, size, cellclass, cellparams, structure,
                          label):
        pass

    @abstractmethod
    def create_projection(self, presynaptic_population,
                          postsynaptic_population, connector, source,
                          target, synapse_dynamics, label, rng):
        pass



from spynnaker7._version import __version__  # NOQA
from spynnaker7._version import __version_name__  # NOQA
from spynnaker7._version import __version_month__  # NOQA
from spynnaker7._version import __version_year__  # NOQA



from pyNN.space import Space

from spynnaker.pyNN.models.neural_projections.connectors. \
    fixed_probability_connector import FixedProbabilityConnector as \
    CommonFixedProbabilityConnector
from pyNN.connectors import FixedProbabilityConnector as \
    PyNNFixedProbabilityConnector


class FixedProbabilityConnector(
        CommonFixedProbabilityConnector, PyNNFixedProbabilityConnector):
    """
    """

    def __init__(
            self, p_connect, weights=0.0, delays=1,
            allow_self_connections=True, space=Space(), safe=True,
            verbose=False):
        """ For each pair of pre-post cells, the connection probability is
         constant.

        :param p_connect: a float between zero and one. Each potential
            connection is created with this probability.
        :param allow_self_connections: if the connector is used to connect a
            Population to itself, this flag determines whether a neuron is
            allowed to connect to itself, or only to other neurons in the
            Population.
        :param safe: if True, check that weights and delays have valid values.
            If False, this check is skipped.
        :param space: a Space object, needed if you wish to specify distance-
            dependent weights or delays - not implemented
        :param verbose:
        """
        CommonFixedProbabilityConnector.__init__(
            self, p_connect=p_connect,
            allow_self_connections=allow_self_connections, safe=safe,
            verbose=verbose)
        self.set_weights_and_delays(weights, delays)
        self.set_space(space)



from spynnaker.pyNN.models.neural_projections.connectors.from_list_connector\
    import FromListConnector as CommonFromListConnector
from spynnaker.pyNN import exceptions
import logging

logger = logging.getLogger(__name__)


class FromListConnector(CommonFromListConnector):
    """ Make connections according to a list.

    :param: conn_list:
        a list of tuples, one tuple for each connection. Each
        tuple should contain::

         (pre_idx, post_idx, weight, delay)

        where pre_idx is the index (i.e. order in the Population,
        not the ID) of the presynaptic neuron, and post_idx is
        the index of the postsynaptic neuron.
    """

    def __init__(self, conn_list, safe=True, verbose=False):
        """
        Creates a new FromListConnector.
        """

        if conn_list is None or len(conn_list) == 0:
            raise exceptions.InvalidParameterType(
                "The connection list for the FromListConnector must contain"
                " at least a list of tuples, each of which should contain:"
                " (pre_idx, post_idx, weight, delay)")

        conn_list, weights, delays, self._extra_conn_data = \
            self._split_conn_list(
                conn_list, ['pre', 'post', 'weight', 'delay'])

        CommonFromListConnector.__init__(self, conn_list, safe, verbose)
        self.set_weights_and_delays(weights, delays)



# common front end imports
import logging

from pyNN.random import NumpyRNG
from pyNN.random import RandomDistribution
from pyNN import __version__ as pynn_version

from spinn_front_end_common.utilities import globals_variables
from spynnaker.pyNN.abstract_spinnaker_common import AbstractSpiNNakerCommon
from spynnaker7.pyNN.models.pynn_population import Population
from spynnaker7.pyNN.models.pynn_projection import Projection
from spynnaker7.pyNN.spynnaker7_simulator_interface import \
    Spynnaker7SimulatorInterface
from spynnaker7.pyNN.utilities.random_stats import \
    RandomStatsScipyImpl, RandomStatsUniformImpl
from spynnaker7.pyNN.utilities.spynnaker7_failed_state \
    import Spynnaker7FailedState
from _version import __version__ as version

# global objects
logger = logging.getLogger(__name__)

# At import time change the default FailedState
globals_variables.set_failed_state(Spynnaker7FailedState())


class Spinnaker(AbstractSpiNNakerCommon, Spynnaker7SimulatorInterface):
    """
    Spinnaker: the main entrance for the spynnaker front end
    """

    def __init__(
            self, host_name=None, timestep=None, min_delay=None,
            max_delay=None, graph_label=None, database_socket_addresses=None,
            n_chips_required=None):
        front_end_versions = [("sPyNNaker7_version", version)]
        front_end_versions.append(("pyNN_version", pynn_version))

        # population holders
        AbstractSpiNNakerCommon.__init__(
            self, database_socket_addresses=database_socket_addresses,
            graph_label=graph_label,
            n_chips_required=n_chips_required, timestep=timestep,
            hostname=host_name, max_delay=max_delay, min_delay=min_delay,
            front_end_versions=front_end_versions)

    def create_population(self, size, cellclass, cellparams, structure, label):
        """ creates a pynn 0.75 population

        :param size: the number of atoms in this population
        :param cellclass: the type of neuron model this pop represents
        :param cellparams: the neuron parameters for this population
        :param structure: something to do with space
        :param label: the human readable label of the population
        :return: a population instance
        """
        return Population(
            size=size, cellclass=cellclass, cellparams=cellparams,
            structure=structure, label=label, spinnaker=self)

    def create_projection(
            self, presynaptic_population, postsynaptic_population, connector,
            source, target, synapse_dynamics, label, rng):
        """

        :param presynaptic_population: source pop this projection goes from
        :param postsynaptic_population: dest pop this projection goes to
        :param connector: the definition of which neurons connect to each other
        :param source:
        :param target: type of projection
        :param synapse_dynamics: plasticity object
        :param label: human readable version of the projection
        :param rng: the random number generator to use on this projection
        :return:
        """
        if label is None:
            label = "Projection {}".format(self._edge_count)
            self._edge_count += 1
        return Projection(
            presynaptic_population=presynaptic_population, label=label,
            postsynaptic_population=postsynaptic_population, rng=rng,
            connector=connector, source=source, target=target,
            synapse_dynamics=synapse_dynamics, spinnaker_control=self,
            machine_time_step=self._machine_time_step,
            timescale_factor=self._time_scale_factor,
            user_max_delay=self.max_delay)

    def get_distribution_to_stats(self):
        return {
            'binomial': RandomStatsScipyImpl("binom"),
            'gamma': RandomStatsScipyImpl("gamma"),
            'exponential': RandomStatsScipyImpl("expon"),
            'lognormal': RandomStatsScipyImpl("lognorm"),
            'normal': RandomStatsScipyImpl("norm"),
            'poisson': RandomStatsScipyImpl("poisson"),
            'uniform': RandomStatsUniformImpl(),
            'randint': RandomStatsScipyImpl("randint"),
            'vonmises': RandomStatsScipyImpl("vonmises")}

    @staticmethod
    def get_random_distribution():
        """
        Depricated use  is_a_pynn_random instead
        """
        return RandomDistribution

    def is_a_pynn_random(self, thing):
        """
        Checks if the thing is a pynn random

        The exact definition of a pynn random can or could change between
        pynn versions so can only be checked against a specific pynn version

        :param thing: any object
        :return: True if this object is a pynn random
        :trype: bol
        """
        return isinstance(thing, RandomDistribution)

    def get_pynn_NumpyRNG(self):
        """
        get specfic PyNN version of NumpyRNG
        :return: NumpyRNG
        :rtype: NumpyRNG
        """
        return NumpyRNG()



from pyNN.recording import files
from spynnaker.pyNN.models.neural_projections.connectors.from_file_connector \
    import FromFileConnector as CommonFromFileConnector


class FromFileConnector(CommonFromFileConnector):
    def __init__(
            self, file, distributed=False, safe=True, verbose=False):
        CommonFromFileConnector.__init__(
            self, file=None, distributed=distributed, safe=safe,
            verbose=verbose)

    def get_reader(self, file):
        """
        get a filereader object using the pynn methods

        :return: A pynn StandardTextFile or similar
        """
        return files.StandardTextFile(file, mode="r")



from pyNN.space import Space

from spynnaker.pyNN.models.neural_projections.connectors\
    .small_world_connector import SmallWorldConnector as \
    CommonSmallWorldConnector


class SmallWorldConnector(CommonSmallWorldConnector):

    def __init__(
            self, degree, rewiring, weights=0.0,
            delays=1, allow_self_connections=True, space=Space(),
            safe=True, verbose=False, n_connections=None):
        CommonSmallWorldConnector.__init__(
            self, degree=degree, rewiring=rewiring,
            allow_self_connections=allow_self_connections, safe=safe,
            verbose=verbose, n_connections=n_connections)
        self.set_weights_and_delays(weights, delays)
        self.set_space(space)



import pyNN
import os

if not pyNN.__version__.startswith("0.7"):
    raise Exception(
        "PyNN version {} found; SpyNNaker 7 requires PyNN version 0.7".format(
            pyNN.__version__))

pynn_path = os.path.dirname(pyNN.__file__)
spinnaker_dir = os.path.join(pynn_path, "spiNNaker")
spinnaker_init = os.path.join(spinnaker_dir, "__init__.py")

if not os.path.exists(spinnaker_dir):
    os.mkdir(spinnaker_dir)
spinn_file = open(spinnaker_init, "w")
spinn_file.write("from spynnaker7.pyNN import *\n")
spinn_file.close()
print "Created", spinnaker_init



import logging

from pyNN.space import Space

from pyNN.random import RandomDistribution

from spynnaker.pyNN.models.neural_projections.connectors. \
    fixed_number_post_connector import FixedNumberPostConnector as \
    CommonFixedNumberPostConnector

logger = logging.getLogger(__file__)


class FixedNumberPostConnector(CommonFixedNumberPostConnector):
    """ pynn connector that puts a fixed number of connections on each of the
        post neurons

    """

    def __init__(
            self, n, weights=0.0, delays=1, allow_self_connections=True,
            space=Space(), safe=True, verbose=False):
        """

        :param n: number of random post-synaptic neurons connected to output
        :type n: int
        :param allow_self_connections: ??????
        :param space: a Space object, needed if you wish to specify distance-
            dependent weights or delays
        :param safe: if True, check that weights and delays have valid values.
            If False, this check is skipped.
        :param verbose: ??????????
        :param `float` weights:
            may either be a float, a !RandomDistribution object, a list/
            1D array with at least as many items as connections to be
            created, or a distance dependence as per a d_expression. Units nA.
        :param `float` delays:  -- as `weights`. If `None`, all synaptic delays
            will be set to the global minimum delay.
        """
        if isinstance(n, RandomDistribution):
            raise NotImplementedError(
                "RandomDistribution is not supported for n in the"
                " implementation of FixedNumberPostConnector on this platform")

        CommonFixedNumberPostConnector.__init__(
            self, n=n, safe=safe, verbose=verbose,
            allow_self_connections=allow_self_connections)
        self.set_weights_and_delays(weights, delays)
        self.set_space(space)



from spynnaker.pyNN.utilities.random_stats import AbstractRandomStats

import scipy.stats


class RandomStatsScipyImpl(AbstractRandomStats):
    """ A Random Statistics object that uses scipy directly
    """

    def __init__(self, distribution_type):
        self._scipy_stats = getattr(scipy.stats, distribution_type)

    def cdf(self, dist, v):
        return self._scipy_stats.cdf(v, *dist.parameters)

    def ppf(self, dist, p):
        return self._scipy_stats.ppf(p, *dist.parameters)

    def mean(self, dist):
        return self._scipy_stats.mean(*dist.parameters)

    def std(self, dist):
        return self._scipy_stats.std(*dist.parameters)

    def var(self, dist):
        return self._scipy_stats.var(*dist.parameters)

    def high(self, dist):
        """ Return the variance of the distribution
        """
        if dist.boundaries is None:
            return None
        else:
            return max(dist.boundaries)

    def low(self, dist):
        """ Return the variance of the distribution
        """
        if dist.boundaries is None:
            return None
        else:
            return min(dist.boundaries)



from pyNN.space import Space
from pyNN.random import RandomDistribution

from spynnaker.pyNN.models.neural_projections.connectors\
    .one_to_one_connector import OneToOneConnector as CommonOneToOneConnector


class OneToOneConnector(CommonOneToOneConnector):
    """
    Where the pre- and postsynaptic populations have the same size, connect
    cell i in the presynaptic pynn_population.py to cell i in the postsynaptic
    pynn_population.py for all i.
    """

    def __init__(
            self, weights=0.0, delays=1, space=Space(), safe=True,
            verbose=False):
        """
        """
        CommonOneToOneConnector.__init__(
            self, safe=safe, verbose=verbose,
            random_number_class=RandomDistribution)
        self.set_weights_and_delays(weights, delays)
        self.set_space(space)



from pyNN.space import Space

from spynnaker.pyNN.models.neural_projections.connectors. \
    all_to_all_connector import AllToAllConnector as CommonAllToAllConnector

import logging

logger = logging.getLogger(__file__)


class AllToAllConnector(CommonAllToAllConnector):
    """ Connects all cells in the presynaptic population to all cells in \
        the postsynaptic population
    """

    def __init__(
            self, weights=0.0, delays=1, allow_self_connections=True,
            space=Space(), safe=True, verbose=None):
        """

        :param allow_self_connections:
            if the connector is used to connect a
            Population to itself, this flag determines whether a neuron is
            allowed to connect to itself, or only to other neurons in the
            Population.
        :type allow_self_connections: bool
        :param space: a Space object, needed if you wish to specify distance-
            dependent weights or delays
        :param safe: if True, check that weights and delays have valid values.
         If False, this check is skipped.
        :param verbose:
    """
        CommonAllToAllConnector.__init__(
            self, allow_self_connections=allow_self_connections, safe=safe,
            verbose=verbose)
        self.set_space(space)
        self.set_weights_and_delays(weights, delays)



import os
from setuptools import setup
from collections import defaultdict

__version__ = None
exec(open("spynnaker7/_version.py").read())
assert __version__

# Build a list of all project modules, as well as supplementary files
main_package = "spynnaker7"
data_extensions = {".aplx", ".boot", ".json", ".xml", ".xsd"}
config_extensions = {".cfg", ".template"}
main_package_dir = os.path.join(os.path.dirname(__file__), main_package)
start = len(main_package_dir)
packages = []
package_data = defaultdict(list)
for dirname, dirnames, filenames in os.walk(main_package_dir):
    if '__init__.py' in filenames:
        package = "{}{}".format(
            main_package, dirname[start:].replace(os.sep, '.'))
        packages.append(package)
    ext_set = set()
    for filename in filenames:
        _, ext = os.path.splitext(filename)
        if ext in data_extensions:
            ext_set.add(ext)
        if ext in config_extensions:
            package = "{}{}".format(
                main_package, dirname[start:].replace(os.sep, '.'))
            package_data[package].append(filename)
    for ext in ext_set:
        package = "{}{}".format(
            main_package, dirname[start:].replace(os.sep, '.'))
        package_data[package].append("*{}".format(ext))

setup(
    name="sPyNNaker7",
    version=__version__,
    description="Extensions of Spinnaker implementation specific for PyNN7",
    url="https://github.com/SpiNNakerManchester/SpyNNaker7",
    packages=packages,
    package_data=package_data,
    install_requires=[
        'sPyNNaker >= 1!4.0.0, < 1!5.0.0',
        'pyNN >= 0.7, < 0.8']
)



# general imports
import logging

# fec imports
from spinn_front_end_common.abstract_models \
    import AbstractSendMeMulticastCommandsVertex
from spinn_front_end_common.utilities import globals_variables
from spinn_front_end_common.utilities.notification_protocol \
    import SocketAddress
from spinn_front_end_common.utility_models import LivePacketGather
from spinn_front_end_common.utilities.notification_protocol \
    import SocketAddress as __SockAddr

# spinnman imports
from spinnman.messages.eieio.eieio_type import EIEIOType

# main
from spynnaker.pyNN.connections \
    import EthernetCommandConnection
from spynnaker.pyNN.connections \
    import EthernetControlConnection
from spynnaker.pyNN.connections \
    import SpynnakerLiveSpikesConnection

# connections
from spynnaker.pyNN.external_devices_models \
    import AbstractEthernetController
from spynnaker.pyNN.external_devices_models \
    import AbstractEthernetSensor
from spynnaker.pyNN.external_devices_models \
    import ArbitraryFPGADevice
from spynnaker.pyNN.external_devices_models \
    import ExternalCochleaDevice
from spynnaker.pyNN.external_devices_models \
    import ExternalFPGARetinaDevice
from spynnaker.pyNN.external_devices_models \
    import MunichMotorDevice
from spynnaker.pyNN.external_devices_models \
    import MunichRetinaDevice

# PushBot Ethernet control
from spynnaker.pyNN.external_devices_models.push_bot.\
    push_bot_control_modules import PushBotLifEthernet

# PushBotSpiNNakerLink control
from spynnaker.pyNN.external_devices_models.push_bot.\
    push_bot_control_modules import PushBotLifSpinnakerLink
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_ethernet import PushBotEthernetLaserDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_ethernet import PushBotEthernetLEDDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_ethernet import PushBotEthernetMotorDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_ethernet import PushBotEthernetRetinaDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_ethernet import PushBotEthernetSpeakerDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_spinnaker_link import PushBotSpiNNakerLinkLaserDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_spinnaker_link import PushBotSpiNNakerLinkLEDDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_spinnaker_link import PushBotSpiNNakerLinkMotorDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_spinnaker_link import PushBotSpiNNakerLinkRetinaDevice
from spynnaker.pyNN.external_devices_models.push_bot \
    .push_bot_spinnaker_link import PushBotSpiNNakerLinkSpeakerDevice

# PushBot Parameters
from spynnaker.pyNN.external_devices_models.push_bot. \
    push_bot_parameters import PushBotLED
from spynnaker.pyNN.external_devices_models.push_bot. \
    push_bot_parameters import PushBotMotor
from spynnaker.pyNN.external_devices_models.push_bot. \
    push_bot_parameters import PushBotRetinaResolution
from spynnaker.pyNN.external_devices_models.push_bot. \
    push_bot_parameters import PushBotLaser
from spynnaker.pyNN.external_devices_models.push_bot. \
    push_bot_parameters import PushBotSpeaker

# push bot retina viewer
from spynnaker.pyNN.external_devices_models.push_bot. \
    push_bot_parameters import PushBotRetinaViewer

# other plugins
from spynnaker.pyNN.protocols \
    import MunichIoSpiNNakerLinkProtocol
from spynnaker.pyNN.spynnaker_external_device_plugin_manager \
    import SpynnakerExternalDevicePluginManager
from spynnaker.pyNN.external_devices_models import ExternalDeviceLifControl

# injector
from spynnaker.pyNN.models.utility_models \
    import SpikeInjector as ExternalDeviceSpikeInjector

# useful functions
add_database_socket_address = \
    SpynnakerExternalDevicePluginManager.add_database_socket_address
activate_live_output_to = \
    SpynnakerExternalDevicePluginManager.activate_live_output_to
activate_live_output_for = \
    SpynnakerExternalDevicePluginManager.activate_live_output_for


logger = logging.getLogger(__name__)

spynnaker_external_devices = SpynnakerExternalDevicePluginManager()


__all__ = [
    "EIEIOType",

    # General Devices
    "ExternalCochleaDevice", "ExternalFPGARetinaDevice",
    "MunichRetinaDevice", "MunichMotorDevice",
    "ArbitraryFPGADevice", "PushBotRetinaViewer",
    "ExternalDeviceLifControl",

    # Pushbot Parameters
    "MunichIoSpiNNakerLinkProtocol",
    "PushBotLaser", "PushBotLED", "PushBotMotor", "PushBotSpeaker",
    "PushBotRetinaResolution",

    # Pushbot Ethernet Parts
    "PushBotLifEthernet", "PushBotEthernetLaserDevice",
    "PushBotEthernetLEDDevice", "PushBotEthernetMotorDevice",
    "PushBotEthernetSpeakerDevice", "PushBotEthernetRetinaDevice",

    # Pushbot SpiNNaker Link Parts
    "PushBotLifSpinnakerLink", "PushBotSpiNNakerLinkLaserDevice",
    "PushBotSpiNNakerLinkLEDDevice", "PushBotSpiNNakerLinkMotorDevice",
    "PushBotSpiNNakerLinkSpeakerDevice", "PushBotSpiNNakerLinkRetinaDevice",

    # Connections
    "SpynnakerLiveSpikesConnection",

    # Provided functions
    "activate_live_output_for",
    "activate_live_output_to",
    "SpikeInjector",
    "register_database_notification_request"

]


def register_database_notification_request(hostname, notify_port, ack_port):
    """ Adds a socket system which is registered with the notification protocol

    :param hostname: ip address of host
    :param notify_port: port for listeing for when database is set up
    :param ack_port: the port for sending back the ack
    :rtype: None
    """
    spynnaker_external_devices.add_socket_address(__SockAddr(
        hostname, notify_port, ack_port))


def EthernetControl(
        n_neurons, params, label=None, local_host=None, local_port=None,
        database_notify_port_num=None, database_ack_port_num=None):
    """ Create a PyNN population that can be included in a network to\
        control an external device which is connected to the host

    :param n_neurons: The number of neurons in the control population
    :param model: Class of a model that implements AbstractEthernetController
    :param params: The parameters of the model
    :param label: An optional label for the population
    :param local_host:\
            The optional local host IP address to listen on for commands
    :param lost_port: The optional local port to listen on for commands
    :param database_ack_port_num:\
            The optional port to which responses to the database notification\
            protocol are to be sent
    :param database_notify_port_num:\
            The optional port to which notifications from the database\
            notification protocol are to be sent
    :return:\
            A pyNN Population which can be used as the target of a Projection.\
            Note that the Population can also be used as the source of a\
            Projection, but it might not send spikes.
    """
    if not issubclass(params['model'], AbstractEthernetController):
        raise Exception(
            "Model must be a subclass of AbstractEthernetController")
    vertex = params['model']
    translator = vertex.get_message_translator()
    ethernet_control_connection = EthernetControlConnection(
        translator, local_host, local_port)
    devices_with_commands = [
        device for device in vertex.get_external_devices()
        if isinstance(device, AbstractSendMeMulticastCommandsVertex)
    ]
    if len(devices_with_commands) > 0:
        ethernet_command_connection = EthernetCommandConnection(
            translator, devices_with_commands, local_host,
            database_notify_port_num)
        add_database_socket_address(
            ethernet_command_connection.local_ip_address,
            ethernet_command_connection.local_port, database_ack_port_num)
    live_packet_gather = LivePacketGather(
        ethernet_control_connection.local_ip_address,
        ethernet_control_connection.local_port,
        message_type=EIEIOType.KEY_PAYLOAD_32_BIT,
        payload_as_time_stamps=False, use_payload_prefix=False)
    spynnaker_external_devices.add_application_vertex(live_packet_gather)
    for partition_id in vertex.get_outgoing_partition_ids():
        spynnaker_external_devices.add_edge(
            vertex, live_packet_gather, partition_id)
    return vertex


def EthernetSensorPopulation(
        model, params, local_host=None,
        database_notify_port_num=None, database_ack_port_num=None):
    """ Create a pyNN population which can be included in a network to\
        receive spikes from a device connected to the host

    :param model: Class of a model that implements AbstractEthernetController
    :param params: The parameters of the model
    :param local_host:\
            The optional local host IP address to listen on for database\
            notification
    :param database_ack_port_num:\
            The optional port to which responses to the database notification\
            protocol are to be sent
    :param database_notify_port_num:\
            The optional port to which notifications from the database\
            notification protocol are to be sent
    :return:\
            A pyNN Population which can be used as the source of a Projection.\
            Note that the Population cannot be used as the target of a\
            Projection.
    """
    if not issubclass(model, AbstractEthernetSensor):
        raise Exception("Model must be a subclass of AbstractEthernetSensor")
    device = model(**params)
    injector_params = dict(device.get_injector_parameters())
    injector_params['notify'] = False

    spike_injector_params = dict(injector_params)
    spike_injector_params['n_neurons'] = device.get_n_neurons()
    spike_injector_params['label'] = device.get_injector_label()

    vertex = SpikeInjector(**injector_params)
    if isinstance(device, AbstractSendMeMulticastCommandsVertex):
        ethernet_command_connection = EthernetCommandConnection(
            device.get_translator(), [device], local_host,
            database_notify_port_num)
        add_database_socket_address(
            ethernet_command_connection.local_ip_address,
            ethernet_command_connection.local_port, database_ack_port_num)
    database_connection = device.get_database_connection()
    if database_connection is not None:
        add_database_socket_address(
            database_connection.local_ip_address,
            database_connection.local_port, database_ack_port_num)
    return vertex


def SpikeInjector(
        n_neurons, label, port=None, notify=True,
        virtual_key=None, database_notify_host=None,
        database_notify_port_num=None, database_ack_port_num=None):
    """ Supports adding a spike injector to the application graph.

    :param n_neurons: the number of neurons the spike injector will emulate
    :type n_neurons: int
    :param notify: allows us to not bother with the database system
    :type notify: bool
    :param label: the label given to the population
    :type label: str
    :param port: the port number used to listen for injections of spikes
    :type port: int
    :param virtual_key: the virtual key used in the routing system
    :type virtual_key: int
    :param database_notify_host: the hostname for the device which is\
            listening to the database notification.
    :type database_notify_host: str
    :param database_ack_port_num: the port number to which a external device\
            will acknowledge that they have finished reading the database and\
            are ready for it to start execution
    :type database_ack_port_num: int
    :param database_notify_port_num: The port number to which a external\
            device will receive the database is ready command
    :type database_notify_port_num: int
    """
    if notify:
        _process_database_socket(
            database_notify_port_num, database_notify_host,
            database_ack_port_num)
    return ExternalDeviceSpikeInjector(
        n_neurons=n_neurons, label=label, port=port, virtual_key=virtual_key)


def _process_database_socket(
        database_notify_port_num, database_notify_host, database_ack_port_num):
    """ code to handle building a database socket address as needed

    :param database_notify_port_num:  the port num where to send the db is \
    written packet.
    :param database_notify_host: the ipaddress of where to send the db is \
    written packet.
    :param database_ack_port_num: the port number to listen on for ack of \
     having read and set them selves up on.
    :rtype: None
    """
    config = globals_variables.get_simulator().config
    if database_notify_port_num is None:
        database_notify_port_num = config.getint("Database", "notify_port")
    if database_notify_host is None:
        database_notify_host = config.get("Database", "notify_hostname")
    if database_ack_port_num is None:
        database_ack_port_num = config.get("Database", "listen_port")
        if database_ack_port_num == "None":
            database_ack_port_num = None

    # build the database socket address used by the notification interface
    database_socket = SocketAddress(
        listen_port=database_ack_port_num,
        notify_host_name=database_notify_host,
        notify_port_no=database_notify_port_num)

    # update socket interface with new demands.
    spynnaker_external_devices.add_socket_address(database_socket)



from spynnaker.pyNN.models.neural_projections.connectors.multapse_connector \
    import MultapseConnector as CommonMultapseConnector


class MultapseConnector(CommonMultapseConnector):
    """
    Create a multapse connector. The size of the source and destination
    populations are obtained when the projection is connected. The number of
    synapses is specified. when instantiated, the required number of synapses
    is created by selecting at random from the source and target populations
    with replacement. Uniform selection probability is assumed.

    :param num_synapses:
        Integer. This is the total number of synapses in the connection.

    """
    def __init__(self, num_synapses, weights=0.0, delays=1, safe=True,
                 verbose=False):
        CommonMultapseConnector.__init__(
            self, num_synapses=num_synapses, safe=safe, verbose=verbose)
        self.set_weights_and_delays(weights, delays)



# spynnaker 8 extra models
from spynnaker.pyNN.models.neuron.builds import IFCondExpStoc
from spynnaker.pyNN.models.neuron.builds import IFCurrDelta
from spynnaker.pyNN.models.neuron.builds import IFCurrExpCa2Adaptive
from spynnaker.pyNN.models.neuron.builds import IzkCondExpBase as IZK_cond_exp
from spynnaker.pyNN.models.neuron.builds import IzkCurrExpBase as IZK_curr_exp
from spynnaker.pyNN.models.neuron.builds\
    import IFCurrDualExpBase as IF_curr_dual_exp

# plastic timing spynnaker 7
from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence \
    import TimingDependenceRecurrent as RecurrentRule
from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence \
    import TimingDependenceSpikeNearestPair as SpikeNearestPair
from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence \
    import TimingDependenceVogels2011 as Vogels2011Rule
from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence \
    import TimingDependencePfisterSpikeTriplet as PfisterSpikeTriplet

# plastic weight spynnaker 7
from spynnaker7.pyNN.models.plasticity_components.weight_dependence \
    .weight_dependence_additive_triplet \
    import WeightDependenceAdditiveTriplet as WeightDependenceAdditiveTriplet

__all__ = [
    # spynnaker 7 models
    'IFCurrDelta', 'IFCurrExpCa2Adaptive', 'IFCondExpStoc',
    'IZK_curr_exp', 'IZK_cond_exp', 'IF_curr_dual_exp',

    # spynnaker 7 plastic stuff
    'WeightDependenceAdditiveTriplet',
    'PfisterSpikeTriplet',
    'SpikeNearestPair',
    'RecurrentRule', 'Vogels2011Rule']



from pyNN.space import Space

from pyNN.random import RandomDistribution

from spynnaker.pyNN.models.neural_projections.connectors. \
    fixed_number_pre_connector import FixedNumberPreConnector as \
    CommonFixedNumberPreConnector

import logging

logger = logging.getLogger(__file__)


class FixedNumberPreConnector(CommonFixedNumberPreConnector):
    """ Connects a fixed number of pre-synaptic neurons selected at random,
        to all post-synaptic neurons
    """

    def __init__(
            self, n, weights=0.0, delays=1, allow_self_connections=True,
            space=Space(), safe=True, verbose=False):
        """
        :param n:
            number of random pre-synaptic neurons connected to output
        :param allow_self_connections:
            if the connector is used to connect a
            Population to itself, this flag determines whether a neuron is
            allowed to connect to itself, or only to other neurons in the
            Population.
        :param `pyNN.Space` space:
            a Space object, needed if you wish to specify distance-
            dependent weights or delays - not implemented
        :param safe: if True, check that weights and delays have valid values.
            If False, this check is skipped.
        :param verbose:
        """

        if isinstance(n, RandomDistribution):
            raise NotImplementedError(
                "RandomDistribution is not supported for n in the"
                " implementation of FixedNumberPreConnector on this platform")

        CommonFixedNumberPreConnector.__init__(
            self, n=n, safe=safe,
            allow_self_connections=allow_self_connections, verbose=verbose)
        self.set_weights_and_delays(weights, delays)
        self.set_space(space)
